@@ -33,9 +33,7 @@
  * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
  */
 
-#ifndef lint
-static char rcsid[] = "$Id: chap.c,v 1.21 1999/03/19 01:20:19 paulus Exp $";
-#endif
+#define RCSID	"$Id: chap.c,v 1.23 1999/08/13 06:46:11 paulus Exp $"
 
 /*
  * TODO:
@@ -53,6 +51,8 @@ static char rcsid[] = "$Id: chap.c,v 1.21 1999/03/19 01:20:19 paulus Exp $";
 #include "chap_ms.h"
 #endif
 
+static const char rcsid[] = RCSID;
+
 /*
  * Command-line options.
  */
