@@ -17,9 +17,7 @@
  * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
  */
 
-#ifndef lint
-static char rcsid[] = "$Id: ipcp.c,v 1.43 1999/03/19 04:23:39 paulus Exp $";
-#endif
+#define RCSID	"$Id: ipcp.c,v 1.50 1999/08/24 05:31:09 paulus Exp $"
 
 /*
  * TODO:
@@ -39,6 +37,8 @@ static char rcsid[] = "$Id: ipcp.c,v 1.43 1999/03/19 04:23:39 paulus Exp $";
 #include "ipcp.h"
 #include "pathnames.h"
 
+static const char rcsid[] = RCSID;
+
 /* global vars */
 ipcp_options ipcp_wantoptions[NUM_PPP];	/* Options that we want to request */
 ipcp_options ipcp_gotoptions[NUM_PPP];	/* Options that peer ack'd */
@@ -48,7 +48,6 @@ ipcp_options ipcp_hisoptions[NUM_PPP];	/* Options that we ack'd */
 bool	disable_defaultip = 0;	/* Don't use hostname for default IP adrs */
 
 /* local vars */
-static int cis_received[NUM_PPP];	/* # Conf-Reqs received */
 static int default_route_set[NUM_PPP];	/* Have set up a default route */
 static int proxy_arp_set[NUM_PPP];	/* Have created proxy arp entry */
 static bool usepeerdns;			/* Ask peer for DNS addrs */
@@ -190,7 +189,7 @@ struct protent ipcp_protent = {
     ip_active_pkt
 };
 
-static void ipcp_clear_addrs __P((int));
+static void ipcp_clear_addrs __P((int, u_int32_t, u_int32_t));
 static void ipcp_script __P((char *));		/* Run an up/down script */
 static void ipcp_script_done __P((void *));
 
@@ -441,16 +440,18 @@ ipcp_resetci(f)
     fsm *f;
 {
     ipcp_options *wo = &ipcp_wantoptions[f->unit];
+    ipcp_options *go = &ipcp_gotoptions[f->unit];
 
     wo->req_addr = wo->neg_addr && ipcp_allowoptions[f->unit].neg_addr;
-    if (wo->ouraddr == 0)
+    if (wo->ouraddr == 0 || disable_defaultip)
 	wo->accept_local = 1;
     if (wo->hisaddr == 0)
 	wo->accept_remote = 1;
     wo->req_dns1 = usepeerdns;	/* Request DNS addresses from the peer */
     wo->req_dns2 = usepeerdns;
-    ipcp_gotoptions[f->unit] = *wo;
-    cis_received[f->unit] = 0;
+    *go = *wo;
+    if (disable_defaultip)
+	go->ouraddr = 0;
 }
 
 
@@ -481,16 +482,11 @@ ipcp_cilen(f)
     }
     if (wo->neg_vj && !go->neg_vj && !go->old_vj) {
 	/* try an older style of VJ negotiation */
-	if (cis_received[f->unit] == 0) {
-	    /* keep trying the new style until we see some CI from the peer */
+	/* use the old style only if the peer did */
+	if (ho->neg_vj && ho->old_vj) {
 	    go->neg_vj = 1;
-	} else {
-	    /* use the old style only if the peer did */
-	    if (ho->neg_vj && ho->old_vj) {
-		go->neg_vj = 1;
-		go->old_vj = 1;
-		go->vj_protocol = ho->vj_protocol;
-	    }
+	    go->old_vj = 1;
+	    go->vj_protocol = ho->vj_protocol;
 	}
     }
 
@@ -548,16 +544,15 @@ ipcp_addci(f, ucp, lenp)
 	    neg = 0; \
     }
 
-#define ADDCIDNS(opt, neg) \
+#define ADDCIDNS(opt, neg, addr) \
     if (neg) { \
-	int addrlen = CILEN_ADDR; \
-	if (len >= addrlen) { \
+	if (len >= CILEN_ADDR) { \
 	    u_int32_t l; \
 	    PUTCHAR(opt, ucp); \
-	    PUTCHAR(addrlen, ucp); \
-	    l = ntohl(0); \
+	    PUTCHAR(CILEN_ADDR, ucp); \
+	    l = ntohl(addr); \
 	    PUTLONG(l, ucp); \
-	    len -= addrlen; \
+	    len -= CILEN_ADDR; \
 	} else \
 	    neg = 0; \
     }
@@ -568,9 +563,9 @@ ipcp_addci(f, ucp, lenp)
     ADDCIVJ(CI_COMPRESSTYPE, go->neg_vj, go->vj_protocol, go->old_vj,
 	    go->maxslotindex, go->cflag);
 
-    ADDCIDNS(CI_MS_DNS1, go->req_dns1);
+    ADDCIDNS(CI_MS_DNS1, go->req_dns1, go->dnsaddr[0]);
 
-    ADDCIDNS(CI_MS_DNS2, go->req_dns2);
+    ADDCIDNS(CI_MS_DNS2, go->req_dns2, go->dnsaddr[1]);
 
     *lenp -= len;
 }
@@ -647,12 +642,31 @@ ipcp_ackci(f, p, len)
 	} \
     }
 
+#define ACKCIDNS(opt, neg, addr) \
+    if (neg) { \
+	u_int32_t l; \
+	if ((len -= CILEN_ADDR) < 0) \
+	    goto bad; \
+	GETCHAR(citype, p); \
+	GETCHAR(cilen, p); \
+	if (cilen != CILEN_ADDR || citype != opt) \
+	    goto bad; \
+	GETLONG(l, p); \
+	cilong = htonl(l); \
+	if (addr != cilong) \
+	    goto bad; \
+    }
+
     ACKCIADDR((go->old_addrs? CI_ADDRS: CI_ADDR), go->neg_addr,
 	      go->old_addrs, go->ouraddr, go->hisaddr);
 
     ACKCIVJ(CI_COMPRESSTYPE, go->neg_vj, go->vj_protocol, go->old_vj,
 	    go->maxslotindex, go->cflag);
 
+    ACKCIDNS(CI_MS_DNS1, go->req_dns1, go->dnsaddr[0]);
+
+    ACKCIDNS(CI_MS_DNS2, go->req_dns2, go->dnsaddr[1]);
+
     /*
      * If there are any remaining CIs, then this packet is bad.
      */
@@ -728,9 +742,6 @@ ipcp_nakci(f, p, len)
         code \
     }
 
-/*
- * Peer returns DNS address in a NAK packet
- */
 #define NAKCIDNS(opt, neg, code) \
     if (go->neg && \
 	((cilen = p[1]) == CILEN_ADDR) && \
@@ -788,12 +799,10 @@ ipcp_nakci(f, p, len)
 
     NAKCIDNS(CI_MS_DNS1, req_dns1,
 	    try.dnsaddr[0] = cidnsaddr;
-	    try.req_dns1 = 0;
 	    );
 
     NAKCIDNS(CI_MS_DNS2, req_dns2,
 	    try.dnsaddr[1] = cidnsaddr;
-	    try.req_dns2 = 0;
 	    );
 
     /*
@@ -932,7 +941,7 @@ ipcp_rejci(f, p, len)
 
 #define REJCIDNS(opt, neg, dnsaddr) \
     if (go->neg && \
-	((cilen = p[1]) == CI_MS_DNS1) && \
+	((cilen = p[1]) == CILEN_ADDR) && \
 	len >= cilen && \
 	p[0] == opt) { \
 	u_int32_t l; \
@@ -1006,8 +1015,6 @@ ipcp_reqci(f, inp, len, reject_if_disagree)
     u_char maxslotindex, cflag;
     int d;
 
-    cis_received[f->unit] = 1;
-
     /*
      * Reset all his options.
      */
@@ -1287,7 +1294,7 @@ ip_check_options()
      * Default our local IP address based on our hostname.
      * If local IP address already given, don't bother.
      */
-    if (wo->ouraddr == 0 && !disable_defaultip) {
+    if (wo->ouraddr == 0) {
 	/*
 	 * Look up our hostname (possibly with domain name appended)
 	 * and take the first IP address as our local IP address.
@@ -1300,17 +1307,6 @@ ip_check_options()
 		wo->ouraddr = local;
 	}
     }
-
-    if (demand && wo->hisaddr == 0) {
-	option_error("remote IP address required for demand-dialling\n");
-	exit(1);
-    }
-#if 0
-    if (demand && wo->accept_remote) {
-	option_error("ipcp-accept-remote is incompatible with demand\n");
-	exit(1);
-    }
-#endif
 }
 
 
@@ -1324,6 +1320,17 @@ ip_demand_conf(u)
 {
     ipcp_options *wo = &ipcp_wantoptions[u];
 
+    if (wo->hisaddr == 0) {
+	/* make up an arbitrary address for the peer */
+	wo->hisaddr = htonl(0x0a707070 + ifunit);
+	wo->accept_remote = 1;
+    }
+    if (wo->ouraddr == 0) {
+	/* make up an arbitrary address for us */
+	wo->ouraddr = htonl(0x0a404040 + ifunit);
+	wo->accept_local = 1;
+	disable_defaultip = 1;	/* don't tell the peer this address */
+    }
     if (!sifaddr(u, wo->ouraddr, wo->hisaddr, GetMask(wo->ouraddr)))
 	return 0;
     if (!sifup(u))
@@ -1407,6 +1414,7 @@ ipcp_up(f)
      */
     if (demand) {
 	if (go->ouraddr != wo->ouraddr || ho->hisaddr != wo->hisaddr) {
+	    ipcp_clear_addrs(f->unit, wo->ouraddr, wo->hisaddr);
 	    if (go->ouraddr != wo->ouraddr) {
 		warn("Local IP address changed to %I", go->ouraddr);
 		script_setenv("OLDIPLOCAL", ip_ntoa(wo->ouraddr));
@@ -1419,7 +1427,6 @@ ipcp_up(f)
 		wo->hisaddr = ho->hisaddr;
 	    } else
 		script_unsetenv("OLDIPREMOTE");
-	    ipcp_clear_addrs(f->unit);
 
 	    /* Set the interface to the new addresses */
 	    mask = GetMask(go->ouraddr);
@@ -1487,6 +1494,8 @@ ipcp_up(f)
 	    if (sifproxyarp(f->unit, ho->hisaddr))
 		proxy_arp_set[f->unit] = 1;
 
+	ipcp_wantoptions[0].ouraddr = go->ouraddr;
+
 	notice("local  IP address %I", go->ouraddr);
 	notice("remote IP address %I", ho->hisaddr);
 	if (go->dnsaddr[0])
@@ -1520,8 +1529,9 @@ ipcp_down(f)
     fsm *f;
 {
     IPCPDEBUG(("ipcp: down"));
-    if (get_ppp_stats(f->unit, &link_stats))
-	link_stats_valid = 1;
+    /* XXX a bit IPv4-centric here, we only need to get the stats
+     * before the interface is marked down. */
+    update_link_stats(f->unit);
     if (ipcp_is_up) {
 	ipcp_is_up = 0;
 	np_down(f->unit, PPP_IP);
@@ -1535,8 +1545,10 @@ ipcp_down(f)
     if (demand) {
 	sifnpmode(f->unit, PPP_IP, NPMODE_QUEUE);
     } else {
+	sifnpmode(f->unit, PPP_IP, NPMODE_DROP);
 	sifdown(f->unit);
-	ipcp_clear_addrs(f->unit);
+	ipcp_clear_addrs(f->unit, ipcp_gotoptions[f->unit].ouraddr,
+			 ipcp_hisoptions[f->unit].hisaddr);
     }
 
     /* Execute the ip-down script */
@@ -1552,13 +1564,11 @@ ipcp_down(f)
  * proxy arp entries, etc.
  */
 static void
-ipcp_clear_addrs(unit)
+ipcp_clear_addrs(unit, ouraddr, hisaddr)
     int unit;
+    u_int32_t ouraddr;  /* local address */
+    u_int32_t hisaddr;  /* remote address */
 {
-    u_int32_t ouraddr, hisaddr;
-
-    ouraddr = ipcp_gotoptions[unit].ouraddr;
-    hisaddr = ipcp_hisoptions[unit].hisaddr;
     if (proxy_arp_set[unit]) {
 	cifproxyarp(unit, hisaddr);
 	proxy_arp_set[unit] = 0;
@@ -1750,7 +1760,8 @@ ipcp_printpkt(p, plen, printer, arg)
 	    case CI_MS_DNS2:
 	        p += 2;
 		GETLONG(cilong, p);
-		printer(arg, "ms-dns %I", htonl(cilong));
+		printer(arg, "ms-dns%d %I", code - CI_MS_DNS1 + 1,
+			htonl(cilong));
 		break;
 	    case CI_MS_WINS1:
 	    case CI_MS_WINS2:
