@@ -16,7 +16,7 @@
  * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
  * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
  *
- * $Id: pppd.h,v 1.37 1999/03/31 05:39:43 paulus Exp $
+ * $Id: pppd.h,v 1.48 1999/09/11 12:08:59 paulus Exp $
  */
 
 /*
@@ -33,13 +33,18 @@
 #include <sys/time.h>		/* for struct timeval */
 #include <net/ppp_defs.h>
 
-#if __STDC__
+#if defined(__STDC__)
 #include <stdarg.h>
 #define __V(x)	x
 #else
 #include <varargs.h>
 #define __V(x)	(va_alist) va_dcl
 #define const
+#define volatile
+#endif
+
+#ifdef INET6
+#include "eui64.h"
 #endif
 
 /*
@@ -96,6 +101,10 @@ typedef struct {
 #define OPT_A2COPY	0x200000 /* addr2 -> second location to rcv value */
 #define OPT_ENABLE	0x400000 /* use *addr2 as enable for option */
 #define OPT_PRIVFIX	0x800000 /* can't be overridden if noauth */
+#define OPT_PREPASS	0x1000000 /* do this opt in pre-pass to find device */
+#define OPT_INITONLY	0x2000000 /* option can only be set in init phase */
+#define OPT_DEVEQUIV	0x4000000 /* equiv to device name */
+#define OPT_DEVNAM	(OPT_PREPASS | OPT_INITONLY | OPT_DEVEQUIV)
 
 #define OPT_VAL(x)	((x) & OPT_VALUE)
 
@@ -121,6 +130,12 @@ struct pppd_stats {
     unsigned int	bytes_out;
 };
 
+/* Used for storing a sequence of words.  Usually malloced. */
+struct wordlist {
+    struct wordlist	*next;
+    char		*word;
+};
+
 /*
  * Global variables.
  */
@@ -144,7 +159,13 @@ extern GIDSET_TYPE groups[NGROUPS_MAX];	/* groups the user is in */
 extern int	ngroups;	/* How many groups valid in groups */
 extern struct pppd_stats link_stats; /* byte/packet counts etc. for link */
 extern int	link_stats_valid; /* set if link_stats is valid */
+extern int	link_connect_time; /* time the link was up for */
 extern int	using_pty;	/* using pty as device (notty or pty opt.) */
+extern int	log_to_fd;	/* logging to this fd as well as syslog */
+extern char	*no_ppp_msg;	/* message to print if ppp not in kernel */
+extern volatile int status;	/* exit status for pppd */
+extern int	devnam_fixed;	/* can no longer change devnam */
+extern int	unsuccess;	/* # unsuccessful connection attempts */
 
 /*
  * Variables set by command-line options.
@@ -161,13 +182,14 @@ extern u_int32_t netmask;	/* IP netmask to set on interface */
 extern bool	lockflag;	/* Create lock file to lock the serial dev */
 extern bool	nodetach;	/* Don't detach from controlling tty */
 extern bool	updetach;	/* Detach from controlling tty when link up */
+extern char	*initializer;	/* Script to initialize physical link */
 extern char	*connector;	/* Script to establish physical link */
 extern char	*disconnector;	/* Script to disestablish physical link */
 extern char	*welcomer;	/* Script to welcome client after connection */
 extern char	*ptycommand;	/* Command to run on other side of pty */
 extern int	maxconnect;	/* Maximum connect time (seconds) */
 extern char	user[MAXNAMELEN];/* Our name for authenticating ourselves */
-extern char	passwd[MAXSECRETLEN];	/* Password for PAP */
+extern char	passwd[MAXSECRETLEN];	/* Password for PAP or CHAP */
 extern bool	auth_required;	/* Peer is required to authenticate */
 extern bool	persist;	/* Reopen link after it goes down */
 extern bool	uselogin;	/* Use /etc/passwd for checking PAP */
@@ -179,36 +201,43 @@ extern char	*ipparam;	/* Extra parameter for ip up/down scripts */
 extern bool	cryptpap;	/* Others' PAP passwords are encrypted */
 extern int	idle_time_limit;/* Shut down link if idle for this long */
 extern int	holdoff;	/* Dead time before restarting */
+extern bool	holdoff_specified; /* true if user gave a holdoff value */
 extern bool	notty;		/* Stdin/out is not a tty */
 extern char	*record_file;	/* File to record chars sent/received */
 extern bool	sync_serial;	/* Device is synchronous serial device */
+extern int	maxfail;	/* Max # of unsuccessful connection attempts */
+extern char	linkname[MAXPATHLEN]; /* logical name for link */
+extern bool	tune_kernel;	/* May alter kernel settings as necessary */
 
 #ifdef PPP_FILTER
 extern struct	bpf_program pass_filter;   /* Filter for pkts to pass */
 extern struct	bpf_program active_filter; /* Filter for link-active pkts */
 #endif
 
-char *current_option;		/* the name of the option being parsed */
-int  privileged_option;		/* set iff the current option came from root */
-char *option_source;		/* string saying where the option came from */
-
 #ifdef MSLANMAN
 extern bool	ms_lanman;	/* Use LanMan password instead of NT */
 				/* Has meaning only with MS-CHAP challenges */
 #endif
 
+extern char *current_option;	/* the name of the option being parsed */
+extern int  privileged_option;	/* set iff the current option came from root */
+extern char *option_source;	/* string saying where the option came from */
+
 /*
  * Values for phase.
  */
 #define PHASE_DEAD		0
 #define PHASE_INITIALIZE	1
-#define PHASE_DORMANT		2
-#define PHASE_ESTABLISH		3
-#define PHASE_AUTHENTICATE	4
-#define PHASE_CALLBACK		5
-#define PHASE_NETWORK		6
-#define PHASE_TERMINATE		7
-#define PHASE_HOLDOFF		8
+#define PHASE_SERIALCONN	2
+#define PHASE_DORMANT		3
+#define PHASE_ESTABLISH		4
+#define PHASE_AUTHENTICATE	5
+#define PHASE_CALLBACK		6
+#define PHASE_NETWORK		7
+#define PHASE_RUNNING		8
+#define PHASE_TERMINATE		9
+#define PHASE_DISCONNECT	10
+#define PHASE_HOLDOFF		11
 
 /*
  * The following struct gives the addresses of procedures to call
@@ -267,16 +296,19 @@ void untimeout __P((void (*func)(void *), void *arg));
 pid_t run_program __P((char *prog, char **args, int must_exist,
 		       void (*done)(void *), void *arg));
 				/* Run program prog with args in child */
-void demuxprotrej __P((int, int));
-				/* Demultiplex a Protocol-Reject */
+void reopen_log __P((void));	/* (re)open the connection to syslog */
+void update_link_stats __P((int)); /* Get stats at link termination */
+void script_setenv __P((char *, char *));	/* set script env var */
+void script_unsetenv __P((char *));		/* unset script env var */
+void new_phase __P((int));	/* signal start of new phase */
+
+/* Procedures exported from utils.c. */
 void log_packet __P((u_char *, int, char *, int));
 				/* Format a packet and log it with syslog */
 void print_string __P((char *, int,  void (*) (void *, char *, ...),
 		void *));	/* Format a string for output */
 int slprintf __P((char *, int, char *, ...));		/* sprintf++ */
 int vslprintf __P((char *, int, char *, va_list));	/* vsprintf++ */
-void script_setenv __P((char *, char *));	/* set script env var */
-void script_unsetenv __P((char *));		/* unset script env var */
 size_t strlcpy __P((char *, const char *, size_t));	/* safe strcpy */
 size_t strlcat __P((char *, const char *, size_t));	/* safe strncpy */
 void dbglog __P((char *, ...));	/* log a debug message */
@@ -285,13 +317,13 @@ void notice __P((char *, ...));	/* log a notice-level message */
 void warn __P((char *, ...));	/* log a warning message */
 void error __P((char *, ...));	/* log an error message */
 void fatal __P((char *, ...));	/* log an error message and die(1) */
-void reopen_log __P((void));	/* (re)open the connection to syslog */
 
 /* Procedures exported from auth.c */
 void link_required __P((int));	  /* we are starting to use the link */
 void link_terminated __P((int));  /* we are finished with the link */
 void link_down __P((int));	  /* the LCP layer has left the Opened state */
 void link_established __P((int)); /* the link is up; authenticate now */
+void start_networks __P((void));  /* start all the network control protos */
 void np_up __P((int, int));	  /* a network protocol has come up */
 void np_down __P((int, int));	  /* a network protocol has gone down */
 void np_finished __P((int, int)); /* a network protocol no longer needs link */
@@ -306,7 +338,7 @@ void auth_withpeer_success __P((int, int));
 void auth_check_options __P((void));
 				/* check authentication options supplied */
 void auth_reset __P((int));	/* check what secrets we have */
-int  check_passwd __P((int, char *, int, char *, int, char **, int *));
+int  check_passwd __P((int, char *, int, char *, int, char **));
 				/* Check peer-supplied username/password */
 int  get_secret __P((int, char *, char *, char *, int *, int));
 				/* get "secret" for chap */
@@ -363,14 +395,20 @@ int  get_ppp_stats __P((int, struct pppd_stats *));
 				/* Return link statistics */
 int  sifvjcomp __P((int, int, int, int));
 				/* Configure VJ TCP header compression */
-int  sifup __P((int));		/* Configure i/f up (for IP) */
+int  sifup __P((int));		/* Configure i/f up for one protocol */
 int  sifnpmode __P((int u, int proto, enum NPmode mode));
 				/* Set mode for handling packets for proto */
-int  sifdown __P((int));	/* Configure i/f down (for IP) */
+int  sifdown __P((int));	/* Configure i/f down for one protocol */
 int  sifaddr __P((int, u_int32_t, u_int32_t, u_int32_t));
-				/* Configure IP addresses for i/f */
+				/* Configure IPv4 addresses for i/f */
 int  cifaddr __P((int, u_int32_t, u_int32_t));
 				/* Reset i/f IP addresses */
+#ifdef INET6
+int  sif6addr __P((int, eui64_t, eui64_t));
+				/* Configure IPv6 addresses for i/f */
+int  cif6addr __P((int, eui64_t, eui64_t));
+				/* Remove an IPv6 address from i/f */
+#endif
 int  sifdefaultroute __P((int, u_int32_t, u_int32_t));
 				/* Create default route through i/f */
 int  cifdefaultroute __P((int, u_int32_t, u_int32_t));
@@ -381,8 +419,8 @@ int  cifproxyarp __P((int, u_int32_t));
 				/* Delete proxy ARP entry for peer */
 u_int32_t GetMask __P((u_int32_t)); /* Get appropriate netmask for address */
 int  lock __P((char *));	/* Create lock file for device */
+int  relock __P((int));		/* Rewrite lock file with new pid */
 void unlock __P((void));	/* Delete previously-created lock file */
-int  daemon __P((int, int));	/* Detach us from terminal session */
 void logwtmp __P((const char *, const char *, const char *));
 				/* Write entry to wtmp file */
 int  get_host_seed __P((void));	/* Get host-dependent random number seed */
@@ -404,14 +442,15 @@ int  options_from_file __P((char *filename, int must_exist, int check_prot,
 				/* Parse options from an options file */
 int  options_from_user __P((void)); /* Parse options from user's .ppprc */
 int  options_for_tty __P((void)); /* Parse options from /etc/ppp/options.tty */
-void scan_args __P((int argc, char **argv));
-				/* Look for tty name in command-line args */
+int  options_from_list __P((struct wordlist *, int privileged));
+				/* Parse options from a wordlist */
 int  getword __P((FILE *f, char *word, int *newlinep, char *filename));
 				/* Read a word from a file */
 void option_error __P((char *fmt, ...));
 				/* Print an error message about an option */
 int int_option __P((char *, int *));
 				/* Simplified number_option for decimal ints */
+void add_options __P((option_t *)); /* Add extra options */
 
 /*
  * This structure is used to store information about certain
@@ -425,11 +464,24 @@ struct option_info {
 };
 
 extern struct option_info devnam_info;
+extern struct option_info initializer_info;
 extern struct option_info connector_info;
 extern struct option_info disconnector_info;
 extern struct option_info welcomer_info;
 extern struct option_info ptycommand_info;
 
+/*
+ * Hooks to enable plugins to change various things.
+ */
+extern int (*new_phase_hook) __P((int));
+extern int (*idle_time_hook) __P((struct ppp_idle *));
+extern int (*holdoff_hook) __P((void));
+extern int (*pap_check_hook) __P((void));
+extern int (*pap_auth_hook) __P((char *user, char *passwd, char **msgp,
+				 struct wordlist **paddrs,
+				 struct wordlist **popts));
+extern int (*pap_passwd_hook) __P((char *user, char *passwd));
+
 /*
  * Inline versions of get/put char/short/long.
  * Pointer is advanced; we assume that both arguments
@@ -489,12 +541,40 @@ extern struct option_info ptycommand_info;
     PUTCHAR(PPP_UI, p); \
     PUTSHORT(t, p); }
 
+/*
+ * Exit status values.
+ */
+#define EXIT_OK			0
+#define EXIT_FATAL_ERROR	1
+#define EXIT_OPTION_ERROR	2
+#define EXIT_NOT_ROOT		3
+#define EXIT_NO_KERNEL_SUPPORT	4
+#define EXIT_USER_REQUEST	5
+#define EXIT_LOCK_FAILED	6
+#define EXIT_OPEN_FAILED	7
+#define EXIT_CONNECT_FAILED	8
+#define EXIT_PTYCMD_FAILED	9
+#define EXIT_NEGOTIATION_FAILED	10
+#define EXIT_PEER_AUTH_FAILED	11
+#define EXIT_IDLE_TIMEOUT	12
+#define EXIT_CONNECT_TIME	13
+#define EXIT_CALLBACK		14
+#define EXIT_PEER_DEAD		15
+#define EXIT_HANGUP		16
+#define EXIT_LOOPBACK		17
+#define EXIT_INIT_FAILED	18
+#define EXIT_AUTH_TOPEER_FAILED	19
 
+/*
+ * Debug macros.  Slightly useful for finding bugs in pppd, not particularly
+ * useful for finding out why your connection isn't being established.
+ */
 #ifdef DEBUGALL
 #define DEBUGMAIN	1
 #define DEBUGFSM	1
 #define DEBUGLCP	1
 #define DEBUGIPCP	1
+#define DEBUGIPV6CP	1
 #define DEBUGUPAP	1
 #define DEBUGCHAP	1
 #endif
@@ -502,7 +582,7 @@ extern struct option_info ptycommand_info;
 #ifndef LOG_PPP			/* we use LOG_LOCAL2 for syslog by default */
 #if defined(DEBUGMAIN) || defined(DEBUGFSM) || defined(DEBUGSYS) \
   || defined(DEBUGLCP) || defined(DEBUGIPCP) || defined(DEBUGUPAP) \
-  || defined(DEBUGCHAP) || defined(DEBUG)
+  || defined(DEBUGCHAP) || defined(DEBUG) || defined(DEBUGIPV6CP)
 #define LOG_PPP LOG_LOCAL2
 #else
 #define LOG_PPP LOG_DAEMON
@@ -539,6 +619,12 @@ extern struct option_info ptycommand_info;
 #define IPCPDEBUG(x)
 #endif
 
+#ifdef DEBUGIPV6CP
+#define IPV6CPDEBUG(x)  if (debug) dbglog x
+#else
+#define IPV6CPDEBUG(x)
+#endif
+
 #ifdef DEBUGUPAP
 #define UPAPDEBUG(x)	if (debug) dbglog x
 #else
