@@ -25,9 +25,7 @@
  * OR MODIFICATIONS.
  */
 
-#ifndef lint
-static char rcsid[] = "$Id: sys-svr4.c,v 1.29 1999/03/22 05:55:39 paulus Exp $";
-#endif
+#define RCSID	"$Id: sys-svr4.c,v 1.34 1999/09/08 01:11:16 masputra Exp $"
 
 #include <limits.h>
 #include <stdio.h>
@@ -60,13 +58,17 @@ static char rcsid[] = "$Id: sys-svr4.c,v 1.29 1999/03/22 05:55:39 paulus Exp $";
 #include <net/ppp_defs.h>
 #include <net/pppio.h>
 #include <netinet/in.h>
+#ifdef SOL2
 #include <sys/tihdr.h>
 #include <sys/tiuser.h>
 #include <inet/common.h>
 #include <inet/mib2.h>
+#endif
 
 #include "pppd.h"
 
+static const char rcsid[] = RCSID;
+
 static int	pppfd;
 static int	fdmuxid = -1;
 static int	ipfd;
@@ -109,6 +111,37 @@ static int dlpi_info_req __P((int));
 static int dlpi_get_reply __P((int, union DL_primitives *, int, int));
 static int strioctl __P((int, int, void *, int, int));
 
+#ifdef SOL2
+/*
+ * sifppa - Sets interface ppa
+ *
+ * without setting the ppa, ip module will return EINVAL upon setting the
+ * interface UP (SIOCSxIFFLAGS). This is because ip module in 2.8 expects
+ * two DLPI_INFO_REQ to be sent down to the driver (below ip) before
+ * IFF_UP can be set. Plumbing the device causes one DLPI_INFO_REQ to
+ * be sent down, and the second DLPI_INFO_REQ is sent upon receiving
+ * IF_UNITSEL (old) or SIOCSLIFNAME (new) ioctls. Such setting of the ppa
+ * is required because the ppp DLPI provider advertises itself as
+ * a DLPI style 2 type, which requires a point of attachment to be
+ * specified. The only way the user can specify a point of attachment
+ * is via SIOCSLIFNAME or IF_UNITSEL.
+ *
+ * Such changes in the behavior of ip module was made to meet new or
+ * evolving standards requirements.
+ *
+ */
+static int
+sifppa(fd, ppa)
+        int fd;
+        int ppa;
+{
+        if (ioctl(fd, IF_UNITSEL, (char *)&ppa) < 0) {
+	        return 0;
+        }
+
+        return 1;
+}
+#endif /* SOL2 */
 
 /*
  * sys_init - System-dependent initialization.
@@ -161,6 +194,18 @@ sys_init()
 	close(ifd);
 	fatal("Can't push IP module: %m");
     }
+
+#ifdef SOL2
+    /*
+     * Assign ppa according to the unit number returned by ppp device
+     * after plumbing is completed above.
+     */
+    if (sifppa(ifd, ifunit) < 0) {
+        close (ifd);
+        fatal("Can't set ppa for unit %d: %m", ifunit);
+    }
+#endif /* SOL2 */
+
 #else
     if (dlpi_attach(ifd, ifunit) < 0 ||
 	dlpi_get_reply(ifd, &reply.prim, DL_OK_ACK, sizeof(reply)) < 0) {
@@ -222,7 +267,7 @@ sys_check_options()
     return 1;
 }
 
-
+#if 0
 /*
  * daemon - Detach us from controlling terminal session.
  */
@@ -246,6 +291,7 @@ daemon(nochdir, noclose)
     }
     return 0;
 }
+#endif
 
 /*
  * ppp_available - check whether the system has any ppp interfaces
@@ -277,11 +323,14 @@ establish_ppp(fd)
 
     /* Push the async hdlc module and the compressor module. */
     tty_npushed = 0;
-    if (ioctl(fd, I_PUSH, "ppp_ahdl") < 0) {
-	error("Couldn't push PPP Async HDLC module: %m");
-	return -1;
+
+    if(!sync_serial) {
+        if (ioctl(fd, I_PUSH, "ppp_ahdl") < 0) {
+            error("Couldn't push PPP Async HDLC module: %m");
+	    return -1;
+        }
+        ++tty_npushed;
     }
-    ++tty_npushed;
     if (kdebugflag & 4) {
 	i = PPPDBG_LOG + PPPDBG_AHDLC;
 	strioctl(pppfd, PPPIO_DEBUG, &i, sizeof(int), 0);
@@ -453,6 +502,9 @@ struct speed {
 #ifdef B57600
     { 57600, B57600 },
 #endif
+#ifdef B76800
+    { 76800, B76800 },
+#endif
 #ifdef B115200
     { 115200, B115200 },
 #endif
@@ -509,7 +561,7 @@ set_up_tty(fd, local)
     struct termiox tiox;
 #endif
 
-    if (tcgetattr(fd, &tios) < 0)
+    if (!sync_serial && tcgetattr(fd, &tios) < 0)
 	fatal("tcgetattr: %m");
 
 #ifndef CRTSCTS
@@ -574,7 +626,7 @@ set_up_tty(fd, local)
 	    fatal("Baud rate for %s is 0; need explicit baud rate", devnam);
     }
 
-    if (tcsetattr(fd, TCSAFLUSH, &tios) < 0)
+    if (!sync_serial && tcsetattr(fd, TCSAFLUSH, &tios) < 0)
 	fatal("tcsetattr: %m");
 
 #ifndef CRTSCTS
@@ -1610,9 +1662,9 @@ logwtmp(line, name, host)
 
     if (name[0] != 0) {
 	/* logging in */
-	strlcpy(utmpx.ut_user, name, sizeof(utmpx.ut_user));
-	strlcpy(utmpx.ut_id, ifname, sizeof(utmpx.ut_id));
-	strlcpy(utmpx.ut_line, line, sizeof(utmpx.ut_line));
+	strncpy(utmpx.ut_user, name, sizeof(utmpx.ut_user));
+	strncpy(utmpx.ut_id, ifname, sizeof(utmpx.ut_id));
+	strncpy(utmpx.ut_line, line, sizeof(utmpx.ut_line));
 	utmpx.ut_pid = getpid();
 	utmpx.ut_type = USER_PROCESS;
     } else {
@@ -1656,6 +1708,7 @@ strioctl(fd, cmd, ptr, ilen, olen)
     return 0;
 }
 
+#if 0
 /*
  * lock - create a lock file for the named lock device
  */
@@ -1733,7 +1786,7 @@ unlock()
 	lock_file[0] = 0;
     }
 }
-
+#endif
 
 /*
  * cifroute - delete a route through the addresses given.
@@ -1775,6 +1828,7 @@ int
 have_route_to(addr)
     u_int32_t addr;
 {
+#ifdef SOL2
     int fd, r, flags, i;
     struct {
 	struct T_optmgmt_req req;
@@ -1865,6 +1919,9 @@ have_route_to(addr)
     }
     close(fd);
     return 0;
+#else
+    return -1;
+#endif /* SOL2 */
 }
 
 /*
